#ifndef ImageRecord_h
#define ImageRecord_h

// $Id: ImageRecord.h,v 1.15 2014/02/10 10:34:29 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "FileRecord.h"
#include "RefPoint.h"

#include <QDomElement>
#include <QDomDocument>
#include <QList>

class ImageRecord: public Counter
{

    public:

    //! list iterator
    typedef QListIterator<ImageRecord> ListIterator;

    //! constructor
    ImageRecord( const File& = File(), int = 0 );

    //! destructor
    virtual~ImageRecord( void )
    {}

    //! constructor
    ImageRecord( const QDomElement& );

    //! write to dom
    QDomElement domElement( QDomDocument& ) const;

    //! equal to operator
    bool operator == (const ImageRecord& other ) const
    {
        return
            file() == other.file() &&
            version() == other.version();
    }

    //! less than operator
    bool operator < (const ImageRecord& other ) const
    {
        if( file() == other.file() ) return version() < other.version();
        else return file() < other.file();
    }

    //!@name accessors
    //@{

    //! file
    const FileRecord& file( void ) const
    { return file_; }

    //! file
    FileRecord& file( void )
    { return file_; }

    //! version
    int version( void ) const
    { return version_; }

    //! reference points
    const RefPoint::List& refPoints( void ) const
    { return refPoints_; }

    //! collection
    const RefPoint::Collection& collection( void ) const
    { return dataSets_; }

    //! data set index
    int dataSetIndex( void ) const
    { return dataSetIndex_; }

    //! flags
    enum Flag
    {
        LogX = 1<<0,
        LogY = 1<<1
    };

    Q_DECLARE_FLAGS( Flags, Flag )

    //! flags
    Flags flags( void ) const
    { return flags_; }

    //! flags
    bool hasFlag( Flag flag ) const
    { return flags_ & flag; }

    //@}

    //!@name modifiers
    //@{

    //! file
    void setFile( const QString& file )
    { file_ = FileRecord( file ); }

    //! file
    void setFile( const FileRecord& record )
    { file_ = record; }

    //! version
    void setVersion( int value )
    {  version_ = value; }

    //! reference points
    void setRefPoints( const RefPoint::List& points )
    { refPoints_ = points; }

    //! set collection
    void setCollection( const RefPoint::Collection& value )
    { dataSets_ = value; }

    //! data set index
    void setDataSetIndex( int value )
    { dataSetIndex_ = value; }

    //! flags
    void setFlags( Flags value )
    { flags_ = value; }

    //! flags
    void setFlag( Flag flag, bool value )
    {
        if( value ) flags_ |= flag;
        else flags_ &= ~flag;
    }

    //@}

    //! used to sort records according to files
    class FileFTor
    {

        public:

        //! predicate
        bool operator() (const ImageRecord& first, const ImageRecord& second) const
        { return first.file().file() < second.file().file(); }

    };

    //! used to retrieve FileRecord with identical filenames
    class SameFileFTor
    {

        public:

        //! constructor
        SameFileFTor( const File& file = File(), int version = -1 ):
            file_( file ),
            version_( version )
        {}

        //! constructor
        SameFileFTor( const ImageRecord& record ):
            file_( record.file().file() ),
            version_( record.version() )
        {}

        //! predicate
        bool operator() (const ImageRecord& record ) const
        { return record.file().file() == file_ && ( version_ < 0 || record.version() == version_ ); }

        //! predicate
        bool operator() (const ImageRecord& first, const ImageRecord& second) const
        { return first.file().file() == second.file().file() && first.version() == second.version(); }

        private:

        //! filename
        File file_;

        //! version
        int version_;

    };

    //! used to retrieve FileRecord with identical filenames
    class FirstOpenFTor
    {

        public:

        //! predicate
        bool operator() (const ImageRecord& first, const ImageRecord& second) const
        { return first.file().time() < second.file().time(); }

    };

    //! used to remove non existing files
    class InvalidFTor
    {
        public:

        //! predicate
        bool operator()( const ImageRecord& record )
        { return !record.file().isValid(); }

    };


    //! shortcut to list of records
    class List: public QList<ImageRecord>, public Counter
    {
        public:

        //! constructor
        List( void ):
            Counter( "ImageRecord::List" )
        {}

        //! copy constructor
        List( const QList<ImageRecord>& other ):
            QList<ImageRecord>( other ),
            Counter( "ImageRecord::List" )
            {}

        //! constructor
        List( const QDomElement& );

        //! destructor
        virtual ~List( void )
        {}

        //! write to dom
        QDomElement domElement( QDomDocument& ) const;

    };

    protected:

    //! add dataSet
    void _addDataSet( const RefPoint::List& points )
    { dataSets_ << points; }

    private:

    //! file record
    FileRecord file_;

    //! version
    int version_;

    //! reference point
    RefPoint::List refPoints_;

    //! reference points
    RefPoint::Collection dataSets_;

    //! data set index
    int dataSetIndex_;

    //! flags
    Flags flags_;

};

Q_DECLARE_OPERATORS_FOR_FLAGS( ImageRecord::Flags )

#endif
