// $Id: RefPointPanel.cpp,v 1.37 2013/11/15 14:50:39 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "RefPointPanel.h"
#include "RefPointPanel.moc"

#include "BaseContextMenu.h"
#include "CustomToolButton.h"
#include "GridLayout.h"
#include "ZoomWidget.h"

#include <QListView>
#include <QStyleOptionViewItemV4>

#include <cmath>

//______________________________________________________
RefPointPanel::RefPointPanel( QWidget* parent ):
    QWidget( parent ),
    Counter( "RefPointPanel" ),
    matrix_( 4, 0.0 )
{

    // some styles require an item view passed to painting method to have proper selection rendered in items
    itemView_ = new QListView( this );
    itemView_->hide();

    QHBoxLayout* mainLayout( new QHBoxLayout() );
    setLayout( mainLayout );
    mainLayout->setMargin(2);
    mainLayout->setSpacing(5);

    // initialize
    for( int index = 0; index < 3; ++index )
    { refPointWidgets_ << 0x0; }

    // layout
    QVBoxLayout* vlayout( new QVBoxLayout() );
    mainLayout->addLayout( vlayout );
    GridLayout* gridLayout( new GridLayout() );
    vlayout->addLayout( gridLayout );
    vlayout->addStretch( 1 );

    gridLayout->setMaxCount( 4 );
    gridLayout->setMargin(0);
    gridLayout->setSpacing(0);

    // initialize ref points
    const QString names[RefPointCount] =
    {
        tr( " First reference:" ),
        tr( " Second reference:" ),
        tr( " Third reference:" )
    };

    // button group
    editActions_ = new QActionGroup( this );
    editActions_->setExclusive( false );

    // button group
    clearActions_ = new QActionGroup( this );
    clearActions_->setExclusive( false );

    for( int index = 0; index < RefPointCount; ++index )
    {

        RefPointWidget* local( new RefPointWidget( this ) );
        refPointWidgets_[index] = local;

        // set name
        local->nameLabel().setText( names[index] );

        // event filters
        local->nameLabel().installEventFilter( this );
        local->globalPointLabel().installEventFilter( this );
        local->localPointLabel().installEventFilter( this );

        // add to layout
        gridLayout->addWidget( &local->nameLabel(), index, 0 );
        gridLayout->addWidget( &local->globalPointLabel(), index, 1  );
        gridLayout->addWidget( &local->localPointLabel(), index, 2  );
        gridLayout->addWidget( &local->toolbar(), index, 3  );

        editActions_->addAction( &local->editAction() );
        clearActions_->addAction( &local->clearAction() );

        connect( local, SIGNAL(modified()), SLOT(_updateTransformation()) );
        connect( local, SIGNAL(modified()), this, SIGNAL(selectionChanged()) );

    }

    QFrame* separator( new QFrame(this) );
    separator->setFrameStyle( QFrame::VLine );
    mainLayout->addWidget( separator );

    // vertical layout for options
    vlayout = new QVBoxLayout();
    vlayout->setMargin(0);
    vlayout->setSpacing(2);
    mainLayout->addLayout( vlayout );

    vlayout->addWidget( zoomWidget_ = new ZoomWidget( this ) );
    vlayout->addWidget( logXCheckBox_ = new QCheckBox( tr( "Use log scale for horizontal axis" ), this ) );
    vlayout->addWidget( logYCheckBox_ = new QCheckBox( tr( "Use log scale for vertical axis" ), this ) );
    vlayout->addStretch();

    connect( logXCheckBox_, SIGNAL(toggled(bool)), SLOT(_updateTransformation()) );
    connect( logYCheckBox_, SIGNAL(toggled(bool)), SLOT(_updateTransformation()) );

    mainLayout->addStretch(1);

    // context menu
    setContextMenuPolicy( Qt::CustomContextMenu );
    connect( this, SIGNAL(customContextMenuRequested(QPoint)), SLOT(_updateContextMenu(QPoint)) );

}


//________________________________________________________________________
int RefPointPanel::firstInvalidIndex( void ) const
{

    // ensure that not valid
    Q_ASSERT( !isValid() );

    for( int index = 0; index < RefPointCount; ++index )
    { if( !refPointWidgets_[index]->isValid() ) return index; }

    return 0;

}

//________________________________________________________________________
RefPoint::PointF RefPointPanel::toLocal( const RefPoint::PointF& global ) const
{
    Q_ASSERT( isValid() );

    const double x( hasLogX() ?
        exp( origin_.x() + matrix_[0]*global.x() + matrix_[1]*global.y() ):
        origin_.x() + matrix_[0]*global.x() + matrix_[1]*global.y() );

    const double y( hasLogY() ?
        exp( origin_.y() + matrix_[2]*global.x() + matrix_[3]*global.y() ):
        origin_.y() + matrix_[2]*global.x() + matrix_[3]*global.y() );

    return RefPoint::PointF( x, y );

}

//________________________________________________________________________
RefPoint::PointF RefPointPanel::toGlobal( const RefPoint::PointF& local ) const
{
    Q_ASSERT( isValid() );

    const double xLocal( hasLogX() ? log(local.x()):local.x() );
    const double yLocal( hasLogY() ? log(local.y()):local.y() );
    const double determinant( matrix_[0]*matrix_[3] - matrix_[2]*matrix_[1] );

    const double x = ( matrix_[3]*(xLocal - origin_.x()) - matrix_[1]*(yLocal - origin_.y()) )/determinant;
    const double y = ( matrix_[0]*(yLocal - origin_.y()) - matrix_[2]*(xLocal - origin_.x()) )/determinant;
    return RefPoint::PointF( x, y );

}

//________________________________________________________________________
RefPoint RefPointPanel::toLocal( const RefPoint& point ) const
{

    RefPoint out( point );
    out.setLocal( toLocal( out.global() ) );

    // statistical error x
    if( out.errorX().mode_ != RefPoint::NoError )
    {

        // plus
        const RefPoint::PointF localPlus( toLocal( out.errorX().globalPlus_ ) );
        double plus( localPlus.x() - out.local().x() );
        if( out.errorX().relative_ ) plus = out.local().x() != 0 ? plus/out.local().x() : 0;
        out.errorX().plus_ = std::abs( plus );

        // minus
        if( out.errorX().mode_ == RefPoint::ErrorSymetric ) out.errorX().minus_ = out.errorX().plus_;
        else {
            const RefPoint::PointF localMinus( toLocal( out.errorX().globalMinus_ ) );
            double minus( localMinus.x() - out.local().x() );
            if( out.errorX().relative_ ) minus = out.local().x() != 0 ? minus/out.local().x() : 0;
            out.errorX().minus_ = std::abs( minus );
        }

    }

    // statistical error y
    if( out.errorY().mode_ != RefPoint::NoError )
    {

        // plus
        const RefPoint::PointF localPlus( toLocal( out.errorY().globalPlus_ ) );
        double plus( localPlus.y() - out.local().y() );
        if( out.errorY().relative_ ) plus = out.local().y() != 0 ? plus/out.local().y() : 0;
        out.errorY().plus_ = std::abs( plus );

        // minus
        if( out.errorY().mode_ == RefPoint::ErrorSymetric ) out.errorY().minus_ = out.errorY().plus_;
        else {

            const RefPoint::PointF localMinus( toLocal( out.errorY().globalMinus_ ) );
            double minus( localMinus.y() - out.local().y() );
            if( out.errorY().relative_ ) minus = out.local().y() != 0 ? minus/out.local().y() : 0;
            out.errorY().minus_ = std::abs( minus );

        }

    }

    // systematic error x
    if( out.errorSystX().mode_ != RefPoint::NoError )
    {

        // plus
        const RefPoint::PointF localPlus( toLocal( out.errorSystX().globalPlus_ ) );
        double plus( localPlus.x() - out.local().x() );
        if( out.errorSystX().relative_ ) plus = out.local().x() != 0 ? plus/out.local().x() : 0;
        out.errorSystX().plus_ = std::abs( plus );

        // minus
        if( out.errorSystX().mode_ == RefPoint::ErrorSymetric ) out.errorSystX().minus_ = out.errorSystX().plus_;
        else {
            const RefPoint::PointF localMinus( toLocal( out.errorSystX().globalMinus_ ) );
            double minus( localMinus.x() - out.local().x() );
            if( out.errorSystX().relative_ ) minus = out.local().x() != 0 ? minus/out.local().x() : 0;
            out.errorSystX().minus_ = std::abs( minus );
        }

    }

    // systelatic error y
    if( out.errorSystY().mode_ != RefPoint::NoError )
    {

        // plus
        const RefPoint::PointF localPlus( toLocal( out.errorSystY().globalPlus_ ) );
        double plus( localPlus.y() - out.local().y() );
        if( out.errorSystY().relative_ ) plus = out.local().y() != 0 ? plus/out.local().y() : 0;
        out.errorSystY().plus_ = std::abs( plus );

        // minus
        if( out.errorSystY().mode_ == RefPoint::ErrorSymetric ) out.errorSystY().minus_ = out.errorSystY().plus_;
        else {

            const RefPoint::PointF localMinus( toLocal( out.errorSystY().globalMinus_ ) );
            double minus( localMinus.y() - out.local().y() );
            if( out.errorSystY().relative_ ) minus = out.local().y() != 0 ? minus/out.local().y() : 0;
            out.errorSystY().minus_ = std::abs( minus );

        }

    }

    return out;

}

//________________________________________________________________________
RefPoint RefPointPanel::toGlobal( const RefPoint& point ) const
{

    RefPoint out( point );
    out.setGlobal( toGlobal( out.local() ) );

    // statistical error x
    if( out.errorX().mode_ != RefPoint::NoError )
    {

        double plus( std::abs( out.errorX().plus_ ) );
        double minus( out.errorX().mode_ == RefPoint::ErrorSymetric ? plus:std::abs( out.errorX().minus_ ) );

        // convert to absolute
        if( out.errorX().relative_ )
        {
            plus *= out.local().x();
            minus *= out.local().x();
        }

        out.errorX().globalPlus_ = toGlobal( out.local() + QPointF( plus, 0 ) );
        out.errorX().globalMinus_ = toGlobal( out.local() - QPointF( minus, 0 ) );

    }

    // statistical error y
    if( out.errorY().mode_ != RefPoint::NoError )
    {

        double plus( std::abs( out.errorY().plus_ ) );
        double minus( out.errorY().mode_ == RefPoint::ErrorSymetric ? plus:std::abs( out.errorY().minus_ ) );

        // convert to absolute
        if( out.errorY().relative_ )
        {
            plus *= out.local().y();
            minus *= out.local().y();
        }

        out.errorY().globalPlus_ = toGlobal( out.local() + QPointF( 0, plus ) );
        out.errorY().globalMinus_ = toGlobal( out.local() - QPointF( 0, minus ) );

    }

    // systematic error x
    if( out.errorSystX().mode_ != RefPoint::NoError )
    {

        double plus( std::abs( out.errorSystX().plus_ ) );
        double minus( out.errorSystX().mode_ == RefPoint::ErrorSymetric ? plus:std::abs( out.errorSystX().minus_ ) );

        // convert to absolute
        if( out.errorSystX().relative_ )
        {
            plus *= out.local().x();
            minus *= out.local().x();
        }

        out.errorSystX().globalPlus_ = toGlobal( out.local() + QPointF( plus, 0 ) );
        out.errorSystX().globalMinus_ = toGlobal( out.local() - QPointF( minus, 0 ) );

    }

    // systematic error y
    if( out.errorSystY().mode_ != RefPoint::NoError )
    {

        double plus( std::abs( out.errorSystY().plus_ ) );
        double minus( out.errorSystY().mode_ == RefPoint::ErrorSymetric ? plus:std::abs( out.errorSystY().minus_ ) );

        // convert to absolute
        if( out.errorSystY().relative_ )
        {
            plus *= out.local().y();
            minus *= out.local().y();
        }

        out.errorSystY().globalPlus_ = toGlobal( out.local() + QPointF( 0, plus ) );
        out.errorSystY().globalMinus_ = toGlobal( out.local() - QPointF( 0, minus ) );

    }

    return out;

}

//________________________________________________________________________
void RefPointPanel::select( const RefPoint& point )
{
    Debug::Throw( "RefPointPanel::select.\n" );
    for( int index = 0; index < RefPointCount; index++ )
    { refPointWidgets_[index]->setSelected( refPointWidgets_[index]->point() == point ); }
}

//________________________________________________________________________
bool RefPointPanel::select( int i )
{

    Debug::Throw( "RefPointPanel::select.\n" );

    if( i < 0 || i >= RefPointCount || refPointWidgets_[i]->isSelected() ) return false;

    for( int index = 0; index < RefPointCount; index++ )
    { refPointWidgets_[index]->setSelected( index == i ); }

    return true;

}

//________________________________________________________________________
void RefPointPanel::clearSelection( void )
{
    Debug::Throw( "RefPointPanel::clearSelection.\n" );
    for( int index = 0; index < RefPointCount; index++ )
    { refPointWidgets_[index]->setSelected( false ); }
}

//________________________________________________________________________
void RefPointPanel::setPoint( int index, const RefPoint& refPoint )
{ refPointWidgets_[index]->setPoint( refPoint ); }

//________________________________________________________________________
RefPointPanel::ErrorCode RefPointPanel::validate( void ) const
{

    // check whether all points are valid
    foreach( RefPointWidget* refPointWidget, refPointWidgets_ )
    { if( !refPointWidget->isValid() ) return Incomplete; }

    foreach( RefPointWidget* refPointWidget, refPointWidgets_ )
    {
        if( hasLogX() && refPointWidget->point().local().x() <= 0 ) return LogXInvalid;
        if( hasLogY() && refPointWidget->point().local().y() <= 0 ) return LogYInvalid;
    }

    if( _globalDeterminant() == 0 ) return GlobalInvalid;
    else if( _localDeterminant() == 0 ) return LocalInvalid;
    else return Valid;

}

//________________________________________________________________________
bool RefPointPanel::remove( const RefPoint& refPoint )
{
    Debug::Throw( "RefPointPanel::remove.\n" );
    int index( this->index( refPoint ) );
    if( index >= 0 )
    {

        clear( index );
        return true;

    } else return false;

}

//________________________________________________________________________
bool RefPointPanel::eventFilter( QObject* object, QEvent* event )
{
    switch( event->type() )
    {

        case QEvent::HoverEnter:
        {
            bool changed( false );
            for( int index = 0; index < RefPointCount; ++index )
            {

                if( refPointWidgets_[index]->hasChild( object ) && refPointWidgets_[index]->isValid() )
                {

                    changed |= refPointWidgets_[index]->setHovered( true );

                } else {

                    changed |= refPointWidgets_[index]->setHovered( false );

                }

            }

            if( changed ) update();
            break;
        }

        case QEvent::HoverLeave:
        {
            bool changed( false );
            for( int index = 0; index < RefPointCount; ++index )
            {

                if( refPointWidgets_[index]->hasChild( object ) )
                { changed |= refPointWidgets_[index]->setHovered( false ); }

            }

            if( changed ) update();
            break;

        }

        case QEvent::MouseButtonPress:
        {
            bool found( false );
            bool changed( false );
            for( int index = 0; index < RefPointCount; ++index )
            {

                if( refPointWidgets_[index]->hasChild( object ) ) found = true;
                if( refPointWidgets_[index]->hasChild( object ) && refPointWidgets_[index]->isValid() )
                {

                    changed |= refPointWidgets_[index]->setSelected( true );

                } else {

                    changed |= refPointWidgets_[index]->setSelected( false );

                }
            }

            if( changed ) emit selectionChanged();
            if( found ) return true;
            break;
        }

        default: break;

    }

    return false;

}

//________________________________________________________________________
void RefPointPanel::_updateTransformation( void )
{
    Debug::Throw() << "RefPointPanel::_updateTransformation - valid: " << isValid() << endl;
    if( isValid() )
    {

        const double dX2 = refPointWidgets_[1]->point().global().x() - refPointWidgets_[0]->point().global().x();
        const double dY2 = refPointWidgets_[1]->point().global().y() - refPointWidgets_[0]->point().global().y();

        const double dX3 = refPointWidgets_[2]->point().global().x() - refPointWidgets_[0]->point().global().x();
        const double dY3 = refPointWidgets_[2]->point().global().y() - refPointWidgets_[0]->point().global().y();

        double dx2(0), dx3(0);
        if( hasLogX() )
        {

            dx2 = std::log( refPointWidgets_[1]->point().local().x()/refPointWidgets_[0]->point().local().x() );
            dx3 = std::log( refPointWidgets_[2]->point().local().x()/refPointWidgets_[0]->point().local().x() );

        } else {

            dx2 = refPointWidgets_[1]->point().local().x() - refPointWidgets_[0]->point().local().x();
            dx3 = refPointWidgets_[2]->point().local().x() - refPointWidgets_[0]->point().local().x();

        }

        double dy2(0), dy3(0);
        if( hasLogY() )
        {

            dy2 = std::log( refPointWidgets_[1]->point().local().y()/refPointWidgets_[0]->point().local().y() );
            dy3 = std::log( refPointWidgets_[2]->point().local().y()/refPointWidgets_[0]->point().local().y() );

        } else {

            dy2 = refPointWidgets_[1]->point().local().y() - refPointWidgets_[0]->point().local().y();
            dy3 = refPointWidgets_[2]->point().local().y() - refPointWidgets_[0]->point().local().y();

        }

        const double det( _globalDeterminant() );
        matrix_[0] = (dx2*dY3 - dx3*dY2)/det; // a00
        matrix_[1] = (dx3*dX2 - dx2*dX3)/det; // a01
        matrix_[2] = (dy2*dY3 - dy3*dY2)/det; // a10
        matrix_[3] = (dy3*dX2 - dy2*dX3)/det; // a11

        // origin
        if( hasLogX() )
        {

            origin_.setX(
                std::log( refPointWidgets_[0]->point().local().x() )
                - refPointWidgets_[0]->point().global().x()*matrix_[0]
                - refPointWidgets_[0]->point().global().y()*matrix_[1] );

        } else {

            origin_.setX(
                refPointWidgets_[0]->point().local().x()
                - refPointWidgets_[0]->point().global().x()*matrix_[0]
                - refPointWidgets_[0]->point().global().y()*matrix_[1] );

        }

        // origin
        if( hasLogY() )
        {
            origin_.setY(
                std::log( refPointWidgets_[0]->point().local().y() )
                - refPointWidgets_[0]->point().global().x()*matrix_[2]
                - refPointWidgets_[0]->point().global().y()*matrix_[3] );

        } else {

            origin_.setY(
                refPointWidgets_[0]->point().local().y()
                - refPointWidgets_[0]->point().global().x()*matrix_[2]
                - refPointWidgets_[0]->point().global().y()*matrix_[3] );
        }

    }

    emit modified();
}

//____________________________________________________________
void RefPointPanel::_updateContextMenu( const QPoint& position )
{

    Debug::Throw( "RefPointPanel::updateErrorsActions.\n" );

    for( int index = 0; index < RefPointCount; index++ )
    {
        if( refPointWidgets_[index]->isSelected() )
        {
            BaseContextMenu menu( this );
            menu.setHideDisabledActions( true );
            menu.addAction( &refPointWidgets_[index]->editAction() );
            menu.addAction( &refPointWidgets_[index]->clearAction() );

            // execute
            menu.exec( mapToGlobal( position ) );
            break;
        }

    }

    return;

}

//________________________________________________________________________
void RefPointPanel::paintEvent( QPaintEvent* event )
{

    if( hasSelection() || hasHover() )
    {

        QPainter painter( this );
        painter.setClipRegion( event->region() );

        for( int index = 0; index < RefPointCount; index++ )
        {
            if( refPointWidgets_[index]->isSelected() || refPointWidgets_[index]->isHovered() )
            {

                QStyleOptionViewItemV4 option;
                option.initFrom( this );
                option.showDecorationSelected = true;
                option.rect =
                    refPointWidgets_[index]->nameLabel().geometry().united(
                    refPointWidgets_[index]->localPointLabel().geometry() );

                if( refPointWidgets_[index]->isSelected() ) option.state |= QStyle::State_Selected;
                if( refPointWidgets_[index]->isHovered() ) option.state |= QStyle::State_MouseOver;
                style()->drawPrimitive( QStyle::PE_PanelItemViewItem, &option, &painter, itemView_ );

            }
        }

    }

    return QWidget::paintEvent( event );

}

//________________________________________________________________________
double RefPointPanel::_globalDeterminant( void ) const
{
    const double dX2 = refPointWidgets_[1]->point().global().x() - refPointWidgets_[0]->point().global().x();
    const double dY2 = refPointWidgets_[1]->point().global().y() - refPointWidgets_[0]->point().global().y();

    const double dX3 = refPointWidgets_[2]->point().global().x() - refPointWidgets_[0]->point().global().x();
    const double dY3 = refPointWidgets_[2]->point().global().y() - refPointWidgets_[0]->point().global().y();

    return dX2*dY3 - dX3*dY2;
}

//________________________________________________________________________
double RefPointPanel::_localDeterminant( void ) const
{
    const double dX2 = refPointWidgets_[1]->point().local().x() - refPointWidgets_[0]->point().local().x();
    const double dY2 = refPointWidgets_[1]->point().local().y() - refPointWidgets_[0]->point().local().y();

    const double dX3 = refPointWidgets_[2]->point().local().x() - refPointWidgets_[0]->point().local().x();
    const double dY3 = refPointWidgets_[2]->point().local().y() - refPointWidgets_[0]->point().local().y();

    return dX2*dY3 - dX3*dY2;
}
