#ifndef UndoCommand_h
#define UndoCommand_h

// $Id: UndoCommand.h,v 1.12 2013/04/29 18:27:11 hupereir Exp $

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "RefPoint.h"
#include "MainWindow.h"

#include <QPointer>
#include <QTime>
#include <QApplication>
#include <QUndoCommand>

class UndoCommand: public QUndoCommand, public Counter
{

    public:

    //! constructor
    UndoCommand( MainWindow* parent ):
        Counter( "UndoCommand" ),
        parent_( parent ),
        time_( QTime::currentTime() )
    {}

    //! destructor
    virtual ~UndoCommand( void )
    {}

    //!@name accessors
    //@{

    //! time
    const QTime& time( void ) const
    { return time_; }

    //@}

    //! time
    void setTime( const QTime& time )
    { time_ = time; }

    protected:

    enum Type
    {
        MarkerAdded,
        MarkerRemoved,
        MarkerChanged,
        MarkersMovedUp,
        MarkersMovedDown,
        RefPointAdded,
        RefPointRemoved,
        RefPointChanged

    };

    //! parent
    QPointer<MainWindow> _parent( void ) const
    { return parent_; }

    //! true if both commands can be merged
    bool _canMerge( const UndoCommand* other ) const
    { return time().msecsTo( other->time() ) < qApp->doubleClickInterval(); }

    private:

    //! parent
    QPointer<MainWindow> parent_;

    //! creation time
    QTime time_;

};

//! base class for marker commands
class MarkerCommand: public UndoCommand
{

    public:

    //! constructor
    MarkerCommand( MainWindow* parent, int index ):
        UndoCommand( parent ),
        dataSetIndex_( index ),
        refPointIndex_( -1 )
    {}

    //! destructor
    virtual ~MarkerCommand( void )
    {}

    protected:

    //! data set
    int _dataSetIndex( void ) const
    { return dataSetIndex_; }

    //! refpoint index
    void _setRefPointIndex( int value )
    { refPointIndex_ = value; }

    //! refpoint index
    int _refPointIndex( void ) const
    { return refPointIndex_; }

    private:

    //! dataset
    int dataSetIndex_;

    //! ref point index
    int refPointIndex_;

};

//! undo command for adding a marker
class MarkerAddedCommand: public MarkerCommand
{

    public:

    //! constructor
    MarkerAddedCommand( MainWindow* parent, int index, const RefPoint& point ):
        MarkerCommand( parent, index ),
        point_( point )
    { setText( QObject::tr( "Add Marker" ) ); }

    //! destructor
    virtual ~MarkerAddedCommand( void )
    {}

    //! id
    int id( void ) const
    { return MarkerAdded; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    private:

    //! source point
    RefPoint point_;

};

//! undo command for removing a marker
class MarkerRemovedCommand: public MarkerCommand
{

    public:

    //! constructor
    MarkerRemovedCommand( MainWindow* parent, int dataSetIndex, const RefPoint& point ):
        MarkerCommand( parent, dataSetIndex ),
        point_( point )
    { setText( QObject::tr( "Remove Marker" ) ); }

    //! constructor
    MarkerRemovedCommand( MainWindow* parent, int dataSetIndex, int refPointIndex, const RefPoint& point ):
        MarkerCommand( parent, dataSetIndex ),
        point_( point )
    {
        _setRefPointIndex( refPointIndex );
        setText( QObject::tr( "Remove Marker" ) );
    }

    //! destructor
    virtual ~MarkerRemovedCommand( void )
    {}

    //! id
    int id( void ) const
    { return MarkerRemoved; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    private:

    //! source point
    RefPoint point_;

};

//! undo command for modifying a marker
class MarkerChangedCommand: public MarkerCommand
{

    public:

    //! constructor
    MarkerChangedCommand( MainWindow* parent, int index, const RefPoint& source, const RefPoint& destination ):
        MarkerCommand( parent, index ),
        source_( source ),
        destination_( destination )
    { setText( QObject::tr( "Edit Marker" ) ); }

    //! destructor
    virtual ~MarkerChangedCommand( void )
    {}

    //! id
    int id( void ) const
    { return MarkerChanged; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    //! merge with
    virtual bool mergeWith( const QUndoCommand* );

    private:

    //! source point
    RefPoint source_;

    //! destination point
    RefPoint destination_;

};

//! base class for marker commands
class MarkersMovedCommand: public UndoCommand
{

    public:

    //! constructor
    MarkersMovedCommand( MainWindow* parent, int index, const RefPoint::List& points ):
        UndoCommand( parent ),
        dataSetIndex_( index ),
        points_( points )
    {}

    //! destructor
    virtual ~MarkersMovedCommand( void )
    {}

    protected:

    //! move up
    void _moveUp( void );

    //! move down
    void _moveDown( void );

    private:

    //! dataset
    int dataSetIndex_;

    //! ref point list
    RefPoint::List points_;

};

//! move markers up
class MarkersMovedUpCommand: public MarkersMovedCommand
{

    public:

    //! constructor
    MarkersMovedUpCommand( MainWindow* parent, int index, const RefPoint::List& points ):
        MarkersMovedCommand( parent, index, points )
    { setText( QObject::tr( "Move Selection Up" ) ); }


    //! destructor
    virtual ~MarkersMovedUpCommand( void )
    {}

    //! id
    int id( void ) const
    { return MarkersMovedUp; }

    //! undo
    virtual void undo( void )
    { _moveDown(); }

    //! undo
    virtual void redo( void )
    { _moveUp(); }

};

//! move markers down
class MarkersMovedDownCommand: public MarkersMovedCommand
{

    public:

    //! constructor
    MarkersMovedDownCommand( MainWindow* parent, int index, const RefPoint::List& points ):
        MarkersMovedCommand( parent, index, points )
    { setText( QObject::tr( "Move Selection Down" ) ); }

    //! destructor
    virtual ~MarkersMovedDownCommand( void )
    {}

    //! id
    int id( void ) const
    { return MarkersMovedDown; }

    //! undo
    virtual void undo( void )
    { _moveUp(); }

    //! undo
    virtual void redo( void )
    { _moveDown(); }

};

//! base class for reference points commands
class RefPointCommand: public UndoCommand
{

    public:

    //! constructor
    RefPointCommand( MainWindow* parent, int index ):
        UndoCommand( parent ),
        index_( index )
    {}

    //! destructor
    virtual ~RefPointCommand( void )
    {}

    protected:

    //! reference point index
    int _index( void ) const
    { return index_; }

    private:

    //! reference point index
    int index_;

};

//! undo command for adding a reference point
class RefPointAddedCommand: public RefPointCommand
{
    public:

    //! constructor
    RefPointAddedCommand( MainWindow* parent, int index, const RefPoint& point ):
        RefPointCommand( parent, index ),
        point_( point )
    { setText( QObject::tr( "Add Reference Point" ) ); }

    //! destructor
    virtual ~RefPointAddedCommand( void )
    {}

    //! id
    int id( void ) const
    { return RefPointAdded; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    private:

    //! point
    RefPoint point_;

};

//! undo command for removing a reference point
class RefPointRemovedCommand: public RefPointCommand
{
    public:

    //! constructor
    RefPointRemovedCommand( MainWindow* parent, int index, const RefPoint& point ):
        RefPointCommand( parent, index ),
        point_( point )
    { setText( QObject::tr( "Remove Reference Point" ) ); }

    //! destructor
    virtual ~RefPointRemovedCommand( void )
    {}

    //! id
    int id( void ) const
    { return RefPointRemoved; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    private:

    //! point
    RefPoint point_;

};

//! undo command for adding a reference point
class RefPointChangedCommand: public RefPointCommand
{
    public:

    //! constructor
    RefPointChangedCommand( MainWindow* parent, int index, const RefPoint& source, const RefPoint& destination ):
        RefPointCommand( parent, index ),
        source_( source ),
        destination_( destination )
    { setText( QObject::tr( "Edit Reference Point" ) ); }

    //! destructor
    virtual ~RefPointChangedCommand( void )
    {}

    //! id
    int id( void ) const
    { return RefPointRemoved; }

    //! undo
    virtual void undo( void );

    //! redo
    virtual void redo( void );

    //! merge with
    virtual bool mergeWith( const QUndoCommand* );

    private:

    //! source
    RefPoint source_;

    //! source
    RefPoint destination_;

};

#endif
